import os

from fsbc.system import System
from fsgs.drivers.gamedriver import Emulator, GameDriver
from fsgs.input.mapper import InputMapper


class DolphinDriver(GameDriver):
    def __init__(self, fsgs):
        super().__init__(fsgs)
        self.emulator = Emulator("dolphin-emu")
        self.emulator.allow_system_emulator = True

    def prepare(self):
        # configure dolphin.ini
        temp_dir = self.temp_dir("dolphin")
        dolphin_config_file = os.path.join(
            temp_dir.path, "user", "Config", "Dolphin.ini"
        )
        if not os.path.exists(os.path.dirname(dolphin_config_file)):
            os.makedirs(os.path.dirname(dolphin_config_file))

        with open(dolphin_config_file, "w", encoding="UTF-8") as f:
            self.configure(f)

        # media options
        rom_path = self.get_game_file()
        self.emulator.args.extend(["--exec=" + rom_path])

    def finish(self):
        pass

    def configure(self, f):
        temp_dir = self.temp_dir("dolphin")

        f.write("[Interface]\n")
        f.write("HideCursor = True\n")
        f.write("[Display]\n")
        f.write("RenderToMain = True\n")
        f.write(
            "FullscreenResolution = {w}x{h}\n".format(
                w=self.screen_size()[0], h=self.screen_size()[1]
            )
        )
        if self.use_fullscreen():
            f.write("Fullscreen = True\n")
        else:
            f.write("Fullscreen = False\n")

        f.write("[Core]\n")
        if System.windows:
            f.write("GFXPlugin = Plugin_VideoDX9.dll\n")

        # Force Interpreter (Cached) core.
        f.write("CPUCore = 0\n")

        self.dolphin_configure_core(f)
        self.dolphin_configure_input()

        # graphics options
        if self.use_vsync():
            vsync = True
        else:
            vsync = False
        if System.windows:
            graphics_config_file = os.path.join(
                temp_dir, "user", "Config", "gfx_dx9.ini"
            )
            with open(graphics_config_file, "w") as f:
                f.write("[Hardware]\n")
                if vsync:
                    f.write("VSync = True\n")
                else:
                    f.write("VSync = False\n")
                f.write("[Settings]\n")
                f.write("DisableFog = True\n")
                f.write("[Hacks]\n")
                f.write("EFBToTextureEnable = True\n")
        else:
            graphics_config_file = os.path.join(
                temp_dir.path, "user", "Config", "gfx_opengl.ini"
            )
            with open(graphics_config_file, "w") as f:
                f.write("[Hardware]\n")
                if vsync:
                    f.write("VSync = True\n")
                else:
                    f.write("VSync = False\n")
                f.write("[Settings]\n")

    def dolphin_configure_core(self, f):
        pass

    def dolphin_configure_input(self):
        pass


class DolphinInputMapper(InputMapper):
    def axis(self, axis, positive):
        dir_str = "+" if positive else "-"
        return "Axis " + str(axis) + dir_str

    def hat(self, hat, direction):
        dir_str = {"left": "W", "right": "E", "up": "N", "down": "S"}[
            direction
        ]
        return "Hat " + str(hat) + " " + dir_str

    def button(self, button):
        return "Button " + str(button)

    # Mouse values
    # elif value.startswith("M/"):
    #    if value == "M/UP": return "Cursor Y-"
    #    if value == "M/DOWN": return "Cursor Y+"
    #    if value == "M/LEFT": return "Cursor X-"
    #    if value == "M/RIGHT": return "Cursor X+"
    #    if value == "M/00": return "Click 0"
    #    if value == "M/01": return "Click 1"
    #    if value == "M/02": return "Click 2"
    #    raise Exception("unknown mouse value " + value)
    # else:
    #    #if fs.windows:
    #    #    return key_mapping[value]
    #    #else:
    #    #    return value
    #    return ""

    def key(self, key):
        if System.windows:
            return key.dinput_name[4:]
        else:
            return key.sdl_name[5:]

    def mouse(self, button, axis, positive):
        if button:
            return "Click " + str(button - 1)
        else:
            if axis == 0:
                return "Cursor X+" if positive else "Cursor X-"
            if axis == 1:
                return "Cursor Y+" if positive else "Cursor Y-"



import json

from fsgs.drivers.mamedriver import MameDriver
from fsgs.option import Option
from fsgs.platform import Platform
from fsgs.platforms.arcade.arcadeplatform import ArcadeLoader

NEOGEO_PLATFORM_ID = "neogeo"
NEOGEO_PLATFORM_NAME = "Neo-Geo"
NEOGEO_MODEL_MVS = "mvs"
NEOGEO_MODEL_MVS_US = "mvs/us"
NEOGEO_MODEL_MVS_JP = "mvs/jp"
NEOGEO_MODEL_AES = "aes"
NEOGEO_MODEL_AES_JP = "aes/jp"
# noinspection SpellCheckingInspection
NEOGEO_ASIA_BIOS = {"neo-epo.bin": "1b3b22092f30c4d1b2c15f04d1670eb1e9fbea07"}
# noinspection SpellCheckingInspection
NEOGEO_JAPAN_BIOS = {"neo-po.bin": "4e4a440cae46f3889d20234aebd7f8d5f522e22c"}
NEOGEO_CONTROLLER = {
    "type": "gamepad",
    "description": "Gamepad",
    "mapping_name": "neogeo",
}


class NeoGeoPlatform(Platform):
    PLATFORM_ID = NEOGEO_PLATFORM_ID
    PLATFORM_NAME = NEOGEO_PLATFORM_NAME

    def __init__(self):
        super().__init__()

    def driver(self, fsgc):
        return NeoGeoDriver(fsgc)

    def loader(self, fsgc):
        return NeoGeoLoader(fsgc)


class NeoGeoLoader(ArcadeLoader):
    pass


class NeoGeoDriver(MameDriver):
    PORTS = [
        {"description": "Input Port 1", "types": [NEOGEO_CONTROLLER]},
        {"description": "Input Port 2", "types": [NEOGEO_CONTROLLER]},
    ]

    def __init__(self, fsgc):
        super().__init__(fsgc)
        self.helper = NeoGeoHelper(self.options)

    def mame_input_mapping(self, _):
        if self.helper.aes():
            return {
                "START": "P#_START",
                "SELECT": "P#_SELECT",
                "UP": "P#_JOYSTICK_UP",
                "DOWN": "P#_JOYSTICK_DOWN",
                "LEFT": "P#_JOYSTICK_LEFT",
                "RIGHT": "P#_JOYSTICK_RIGHT",
                "A": "P#_BUTTON1",
                "B": "P#_BUTTON2",
                "C": "P#_BUTTON3",
                "D": "P#_BUTTON4",
            }
        else:
            return {
                "START": "START#",
                "SELECT": "COIN#",
                "UP": "P#_JOYSTICK_UP",
                "DOWN": "P#_JOYSTICK_DOWN",
                "LEFT": "P#_JOYSTICK_LEFT",
                "RIGHT": "P#_JOYSTICK_RIGHT",
                "A": "P#_BUTTON1",
                "B": "P#_BUTTON2",
                "C": "P#_BUTTON3",
                "D": "P#_BUTTON4",
            }

    def mame_romset(self):
        romset = {}
        for entry in json.loads(self.options["file_list"]):
            romset[entry["name"]] = entry["sha1"]

        if self.helper.model() == NEOGEO_MODEL_MVS:
            romset.update(
                {"sp-s2.sp1": "4f5ed7105b7128794654ce82b51723e16e389543"}
            )
        elif self.helper.model() == NEOGEO_MODEL_MVS_JP:
            # noinspection SpellCheckingInspection
            romset.update(
                {"vs-bios.rom": "ecf01eda815909f1facec62abf3594eaa8d11075"}
            )
        elif self.helper.model() == NEOGEO_MODEL_MVS_US:
            romset.update(
                {"sp-u2.sp1": "5c6bba07d2ec8ac95776aa3511109f5e1e2e92eb"}
            )
        elif self.helper.model() == NEOGEO_MODEL_AES:
            romset.update(NEOGEO_ASIA_BIOS)
        elif self.helper.model() == NEOGEO_MODEL_AES_JP:
            romset.update(NEOGEO_JAPAN_BIOS)

        if self.helper.aes():
            machine = "aes"
            # noinspection SpellCheckingInspection
            romset.update(
                {"000-lo.lo": "5992277debadeb64d1c1c64b0a92d9293eaf7e4a"}
            )
        else:
            machine = "neogeo"
            # noinspection SpellCheckingInspection
            romset.update(
                {
                    "000-lo.lo": "5992277debadeb64d1c1c64b0a92d9293eaf7e4a",
                    "sfix.sfix": "fd4a618cdcdbf849374f0a50dd8efe9dbab706c3",
                    "sm1.sm1": "42f9d7ddd6c0931fd64226a60dc73602b2819dcf",
                }
            )

        return machine, romset

    def prepare(self):
        super().prepare()
        self.create_mame_layout()

        self.install_mame_hash_file("neogeo.xml")
        self.emulator.args.append(self.options["mame_rom_set"])
        if self.helper.model() == NEOGEO_MODEL_AES_JP:
            self.emulator.args.extend(["-bios", "japan"])
        elif self.helper.model() == NEOGEO_MODEL_MVS_JP:
            self.emulator.args.extend(["-bios", "japan"])
        elif self.helper.model() == NEOGEO_MODEL_MVS_US:
            self.emulator.args.extend(["-bios", "us"])


class NeoGeoHelper:
    def __init__(self, options):
        self.options = options

    def model(self):
        if self.options[Option.NEOGEO_MODEL] == NEOGEO_MODEL_AES:
            return NEOGEO_MODEL_AES
        if self.options[Option.NEOGEO_MODEL] == NEOGEO_MODEL_AES_JP:
            return NEOGEO_MODEL_AES_JP
        if self.options[Option.NEOGEO_MODEL] == NEOGEO_MODEL_MVS_US:
            return NEOGEO_MODEL_MVS_US
        if self.options[Option.NEOGEO_MODEL] == NEOGEO_MODEL_MVS_JP:
            return NEOGEO_MODEL_MVS_JP
        return NEOGEO_MODEL_MVS

    def aes(self):
        return self.model() in [NEOGEO_MODEL_AES, NEOGEO_MODEL_AES_JP]



title = "FS-UAE Launcher"
name = "fs-uae-launcher"
py_name = "fs_uae_launcher"
tar_name = "fs-uae-launcher"
version = "9.8.7dummy"
author = "Frode Solheim"
author_email = "frode@fs-uae.net"
package_map = {
    "amitools": ".",
    "arcade": ".",
    "fsbc": ".",
    "fsboot": ".",
    "fsgs": ".",
    "fspy": ".",
    "fstd": ".",
    "fsui": ".",
    "launcher": ".",
    "oyoyo": ".",
    "workspace": ".",
}
packages = sorted(package_map.keys())
scripts = ["fs-uae-launcher"]



class TextRenderer(object):
    def __init__(self, font):
        self.font = font

    def render_text(self, text, color):
        data, size = self.font.render(text, True, color)

        # swap channels

        # a = numpy.fromstring(data, dtype=numpy.uint8)
        # temp = numpy.array(a[0::4])
        # a[0::4] = numpy.array(a[2::4])
        # a[2::4] = temp
        # return a.tostring(), size

        # edit: no need to swap channels when using GL_BGRA for pixel format

        return data, size

    def get_size(self, text):
        return self.font.rendered_size(text)



from fsgs.network import openretro_url_prefix
from fsui import Image

from ...launcher_config import LauncherConfig
from .WebButton import WebButton


class EditButton(WebButton):
    def __init__(self, parent):
        icon = Image("launcher:res/16x16/pencil.png")
        WebButton.__init__(self, parent, icon)

    def get_url(self):
        variant_uuid = LauncherConfig.get("variant_uuid", "")
        if not variant_uuid:
            return
        return "{0}/game/{1}/edit#{1}".format(
            openretro_url_prefix(), variant_uuid
        )



from __future__ import absolute_import
from __future__ import print_function

from .Block import Block

class BitmapBlock(Block):
  def __init__(self, blkdev, blk_num):
    Block.__init__(self, blkdev, blk_num, chk_loc=0)
  
  def set(self, data):
    self._set_data(data)
    self._read()
  
  def create(self):
    self._create_data()
  
  def read(self):
    self._read_data()
    self._read()
  
  def _read(self):
    Block.read(self)
    if not self.valid:
      return False
    self.valid = True
    return True
    
  def get_bitmap_data(self):
    return self.data[4:]
  
  def set_bitmap_data(self, data):
    self.data[4:] = data
  
  def dump(self):
    Block.dump(self,"Bitmap")



#
# def main(main_window):
#     from arcade.glui.settings import Settings
#
#     try:
#         import arcade.glui.window
#         arcade.glui.window.main_window = main_window
#         arcade.glui.window.show()
#     except Exception as e:
#         import traceback
#         print("\n" + "-" * 79 + "\n" + "EXCEPTION")
#         traceback.print_exc()
#         show_error(repr(e), traceback.format_exc(e))
#
#     print(" --- arcade.glui.main is done ---")



import os

from PIL import Image

for item in os.listdir("."):
    if not item.endswith(".png"):
        continue
    im = Image.open(item)
    if im.mode != "RGB":
        im = im.convert("RGB")
        im.save(item)



import fsui


class Column(fsui.VerticalLayout):
    def __init__(self, parent=None, padding=0):
        super().__init__(padding)
        if parent is not None:
            if hasattr(parent, "set_layout"):
                parent.set_layout(self)
            else:
                parent.layout = self

    def add(
        self,
        element,
        spacing=0,
        expand=False,
        fill=True,
        valign=0.5,
        margin=0,
        left=None,
        right=None,
        top=None,
        bottom=None,
    ):
        super().add(
            element,
            spacing=spacing,
            expand=expand,
            fill=fill,
            valign=valign,
            margin=margin,
            margin_left=left,
            margin_right=right,
            margin_top=top,
            margin_bottom=bottom,
        )

    def expand(self):
        return self.spacer(0, expand=True)

    def row(self, **kwargs):
        row = Row()
        self.add(row, **kwargs)
        return row

    def set_min_size(self, size):
        self.min_width = size[0]
        self.min_height = size[1]

    def spacer(self, size, expand=False):
        self.add_spacer(1, size, expand=expand)


class Row(fsui.HorizontalLayout):
    def __init__(self, parent=None, padding=0):
        super().__init__(padding)
        if parent is not None:
            if hasattr(parent, "set_layout"):
                parent.set_layout(self)
            else:
                parent.layout = self

    def add(
        self,
        element,
        spacing=0,
        expand=False,
        fill=True,
        valign=0.5,
        margin=0,
        left=None,
        right=None,
        top=None,
        bottom=None,
    ):
        super().add(
            element,
            spacing=spacing,
            expand=expand,
            fill=fill,
            valign=valign,
            margin=0,
            margin_left=left,
            margin_right=right,
            margin_top=top,
            margin_bottom=bottom,
        )

    def column(self, **kwargs):
        column = Column()
        self.add(column, **kwargs)
        return column

    def expand(self):
        return self.spacer(0, expand=True)

    def set_min_size(self, size):
        self.min_width = size[0]
        self.min_height = size[1]

    def spacer(self, size, expand=False):
        self.add_spacer(size, 1, expand=expand)



from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals


try:
    # FIXME: this is a hack, there should not be a dependency here
    from launcher.i18n import gettext as _gettext
except ImportError:

    def _gettext(message):
        return message


def gettext(message):
    return _gettext(message)



import fsui
from launcher.cd_manager import CDManager
from launcher.floppy_manager import FloppyManager
from launcher.i18n import gettext
from launcher.option import Option
from launcher.ui.behaviors.platformbehavior import (
    AMIGA_PLATFORMS,
    CDEnableBehavior,
    FloppyEnableBehavior,
)
from launcher.ui.floppyselector import FloppySelector
from launcher.ui.options import ConfigWidgetFactory


class FloppiesGroup(fsui.Group):
    FLOPPY_MODE = FloppySelector.FLOPPY_MODE
    CD_MODE = FloppySelector.CD_MODE
    TAPE_MODE = FloppySelector.TAPE_MODE
    CARTRIDGE_MODE = FloppySelector.CARTRIDGE_MODE

    def __init__(self, parent, drives=2, cd_mode=False, removable_media=False):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        self.cd_mode = cd_mode
        self.num_drives = drives

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)

        self.mode = self.FLOPPY_MODE
        if cd_mode:
            self.mode = self.CD_MODE

        if self.mode == self.CD_MODE:
            title = gettext("CD-ROM Drive")
            drive_count_option = Option.CDROM_DRIVE_COUNT
            behavior_class = CDEnableBehavior
        elif self.mode == self.TAPE_MODE:
            title = gettext("Tape Drive")
            drive_count_option = None
            behavior_class = None
        elif self.mode == self.CARTRIDGE_MODE:
            title = gettext("Cartridge")
            drive_count_option = None
            behavior_class = None
        else:
            title = gettext("Floppy Drives")
            drive_count_option = Option.FLOPPY_DRIVE_COUNT
            behavior_class = FloppyEnableBehavior

        if removable_media:
            # Removable media group will change type dynamically
            behavior_class = None

        self.label = fsui.HeadingLabel(self, title)
        hori_layout.add(self.label, margin=10)
        hori_layout.add_spacer(0, expand=True)

        if drive_count_option and not removable_media:
            # FIXME: Drive count option does not work on the main page when
            # changing to CD mode. Workaround for now is to not include it.
            hori_layout.add(
                ConfigWidgetFactory().create(
                    self,
                    drive_count_option,
                    text=gettext("Drive Count"),
                    platforms=AMIGA_PLATFORMS,
                ),
                fill=True,
                margin_right=20,
            )

        self.multi_select_button = fsui.Button(
            self, gettext("Multi-Select...")
        )
        if self.cd_mode:
            self.multi_select_button.set_tooltip(
                gettext("Add Multiple CD-ROMs at Once")
            )
        else:
            self.multi_select_button.set_tooltip(
                gettext("Add Multiple Floppies at Once")
            )
        if behavior_class:
            behavior_class(self.multi_select_button)
        self.multi_select_button.activated.connect(self.on_multi_select_button)

        hori_layout.add(self.multi_select_button, margin_right=10)

        self.layout.add_spacer(0)

        self.selectors = []
        for i in range(drives):
            selector = FloppySelector(parent, i, show_path=not removable_media)
            if behavior_class:
                behavior_class(selector)
            selector.set_mode(self.mode)
            self.selectors.append(selector)
            self.layout.add(selector, fill=True, margin=10, margin_bottom=0)

    def on_multi_select_button(self):
        if self.cd_mode:
            CDManager.multi_select(self.get_window())
        else:
            FloppyManager.multi_select(self.get_window())

    def update_heading_label(self):
        if self.mode == self.CD_MODE:
            if self.num_drives > 1:
                self.label.set_text(gettext("CD-ROM Drives"))
            else:
                self.label.set_text(gettext("CD-ROM Drive"))
        elif self.mode == self.TAPE_MODE:
            self.label.set_text(gettext("Tape Drive"))
        elif self.mode == self.CARTRIDGE_MODE:
            self.label.set_text(gettext("Cartridge"))
        else:
            self.label.set_text(gettext("Floppy Drives"))
        # Need to update the layout to account for label widget size change.
        self.layout.update()



import hashlib
import os
import traceback

from fsbc.paths import Paths
from fsgs.amiga.rommanager import ROMManager
from fsgs.archive import Archive, archive_extensions
from fsgs.filedatabase import FileDatabase

from .i18n import gettext


class FileScanner(object):
    def __init__(
        self,
        paths,
        purge_other_dirs,
        on_status=None,
        on_rom_found=None,
        stop_check=None,
    ):
        self.paths = paths

        self.purge_other_dirs = purge_other_dirs
        self.database_file_ids = []

        self.on_status = on_status
        self.on_rom_found = on_rom_found
        self._stop_check = stop_check
        self.scan_count = 0
        self.files_added = 0
        self.bytes_added = 0

        self.extensions = set()

        # This will add .zip, .rp9 and .lha, .7z (if extensions are present)
        self.extensions.update(archive_extensions)

        self.extensions.add(".rom")
        self.extensions.add(".adf")
        self.extensions.add(".ipf")
        self.extensions.add(".dms")
        self.extensions.add(".bin")
        self.extensions.add(".cue")
        self.extensions.add(".iso")
        self.extensions.add(".wav")
        self.extensions.add(".fs-uae")

        if True:
            self.extensions.add(".bin")

            # Amstrad CPC
            self.extensions.add(".dsk")

            # Arcade
            self.extensions.add(".chd")

            # Atari 2600
            self.extensions.add(".a26")

            # Atari 5200
            self.extensions.add(".a52")

            # Atari 7800
            self.extensions.add(".a78")

            # Atari ST
            self.extensions.add(".st")  # disk images
            self.extensions.add(".img")  # bios images
            self.extensions.add(".bin")  # bios images

            # Commodore 64

            # DOS / GOG
            # self.extensions.add(".mp3")
            # self.extensions.add(".ogg")

            # Game Boy
            self.extensions.add(".gb")

            # Game Boy Advance
            self.extensions.add(".gba")

            # Game Boy Color
            self.extensions.add(".gbc")

            # Master System
            self.extensions.add(".sms")

            # Mega Drive
            self.extensions.add(".md")
            self.extensions.add(".bin")
            self.extensions.add(".gen")
            # FIXME: Is .smd a common extension?
            self.extensions.add(".smd")

            # Nintendo
            self.extensions.add(".nes")

            # Nintendo 64
            self.extensions.add(".n64")
            self.extensions.add(".v64")
            self.extensions.add(".z64")

            # Nintendo DS
            self.extensions.add(".nds")

            # Super Nintendo
            self.extensions.add(".sfc")
            self.extensions.add(".smc")

            # TurboGrafx-CD system ROM
            self.extensions.add(".pce")

            # ZX Spectrum
            self.extensions.add(".dsk")
            self.extensions.add(".tap")
            self.extensions.add(".tzx")
            self.extensions.add(".z80")

    def stop_check(self):
        if self._stop_check:
            return self._stop_check()

    def set_status(self, title, status):
        if self.on_status:
            self.on_status((title, status))

    def get_scan_dirs(self):
        return self.paths

    def purge_file_ids(self, file_ids):
        self.set_status(
            gettext("Scanning files"), gettext("Purging old entries...")
        )
        database = FileDatabase.get_instance()
        for file_id in file_ids:
            database.delete_file(id=file_id)

    def scan(self):
        self.set_status(gettext("Scanning files"), gettext("Scanning files"))
        file_database = FileDatabase.get_instance()
        # database.clear()
        scan_dirs = self.get_scan_dirs()

        if self.purge_other_dirs:
            all_database_file_ids = file_database.get_file_ids()
        else:
            all_database_file_ids = None

        for dir_ in scan_dirs:
            if not os.path.exists(dir_):
                print("[FILES] Scanner: Directory does not exist:", dir_)
                continue
            # this is important to make sure the database is portable across
            # operating systems
            dir_ = Paths.get_real_case(dir_)

            self.database_file_ids = file_database.get_file_hierarchy_ids(dir_)
            if self.purge_other_dirs:
                all_database_file_ids.difference_update(self.database_file_ids)

            self.scan_dir(file_database, dir_)

            print("Remaining files:", self.database_file_ids)
            self.purge_file_ids(self.database_file_ids)

            self.set_status(
                gettext("Scanning files"), gettext("Committing data...")
            )
            # update last_file_insert and last_file_delete
            file_database.update_last_event_stamps()
            print("[FILES] FileScanner.scan - committing data")
            file_database.commit()

        if self.purge_other_dirs:
            self.purge_file_ids(all_database_file_ids)

        if self.stop_check():
            file_database.rollback()
        else:
            self.set_status(
                gettext("Scanning files"), gettext("Committing data...")
            )
            print("[FILES] FileScanner.scan - committing data")
            file_database.commit()

    def scan_dir(self, file_database, dir_, all_files=False):
        if not os.path.exists(dir_):
            return
        dir_content = os.listdir(dir_)
        if not all_files:
            for name in dir_content:
                if not check_valid_name(name):
                    continue
                l_name = name.lower()
                if l_name.endswith(".slave") or l_name.endswith(".slav"):
                    all_files = True
                    break

        for name in sorted(dir_content):
            if not check_valid_name(name):
                continue
            if self.stop_check():
                return
            if name in [".git", "Cache", "Save States"]:
                continue

            path = os.path.join(dir_, name)
            if os.path.isdir(path):
                self.scan_dir(file_database, path, all_files=all_files)
                continue
            if not all_files:
                dummy, ext = os.path.splitext(path)
                ext = ext.lower()
                if ext not in self.extensions:
                    continue
            try:
                self.scan_file(file_database, path)
            except Exception:
                traceback.print_exc()

    def scan_file(self, file_database, path):
        name = os.path.basename(path)
        # path = os.path.normcase(os.path.normpath(path))

        self.scan_count += 1
        self.set_status(
            gettext("Scanning files ({count} scanned)").format(
                count=self.scan_count
            ),
            name,
        )

        try:
            st = os.stat(path)
        except Exception:
            print("[FILES] WARNING: Error stat-ing file", repr(path))
            return
        size = st.st_size
        mtime = int(st.st_mtime)

        result = file_database.find_file(path=path)
        if result["path"]:
            if size == result["size"] and mtime == result["mtime"]:
                # We've already got this file indexed.
                self.database_file_ids.remove(result["id"])
                return

        archive = Archive(path)
        file_id = self.scan_archive_stream(
            file_database, archive, path, name, size, mtime
        )
        for p in archive.list_files():
            if p.endswith("/"):
                # don't index archive directory entries
                continue
            # print(p)
            if self.stop_check():
                return
            # n = p.replace("\\", "/").replace("|", "/").split("/")[-1]
            n = os.path.basename(p)
            self.scan_count += 1
            self.scan_archive_stream(
                file_database, archive, p, n, size, mtime, parent=file_id
            )
        if self.stop_check():
            return

    def scan_archive_stream(
        self, database, archive, path, name, size, mtime, parent=None
    ):
        self.set_status(
            gettext("Scanning files ({count} scanned)").format(
                count=self.scan_count
            ),
            name,
        )

        f = None
        sha1 = None
        raw_sha1_obj = hashlib.sha1()
        filter_sha1_obj = hashlib.sha1()
        filter_name = ""
        filter_size = 0

        base_name, ext = os.path.splitext(name)
        ext = ext.lower()
        if ext == ".nes":
            # FIXME: NES header hack. Would be better to add a proper notion of
            # file filters to the file database.
            # FIXME: This will confuse some functionality, such as the
            # Locker uploader or other tools expecting on-disk data to match
            # the database checksum (this also applies to the Cloanto ROM
            # hack). Should be done properly.
            f = archive.open(path)
            data = f.read(16)
            if len(data) == 16 and data.startswith(b"NES\x1a"):
                print("Stripping iNES header for", path)
                filter_name = "Skip(16)"
            raw_sha1_obj.update(data)
        elif ext == ".a78":
            # FIXME: Check if 128 is a fixed or variable number of bytes
            f = archive.open(path)
            data = f.read(128)
            if len(data) == 128 and data[1:10] == b"ATARI7800":
                print("Stripping A78 header for", path)
                filter_name = "Skip(128)"
            raw_sha1_obj.update(data)
        elif ext == ".smc":
            f = archive.open(path)
            data = f.read(512)
            if len(data) == 512 and all(map(lambda x: x == 0, data[48:])):
                print("Stripping SMC header for", path)
                filter_name = "Skip(512)"
            raw_sha1_obj.update(data)
        elif ext in [".v64", ".n64"]:
            filter_name = "ByteSwapWords"

        def is_sha1(name):
            if not len(name) == 40:
                return False
            name = name.lower()
            for c in name:
                if c not in "0123456789abcdef":
                    return False
            return True

        if filter_name:
            # We have a filter, so we must calculate checksum of filtered contents
            pass
        else:
            # Try to see if we can deduce the checksum of the contained file.
            # Supports symlinks to compressed files.
            real_path = os.path.realpath(archive.path)
            real_name = os.path.basename(real_path)
            real_name = real_name.lower()
            real_name, real_ext = os.path.splitext(real_name)
            if real_ext in [".xz", ".gz"]:
                if is_sha1(real_name):
                    if parent is not None:
                        # We assume this is the correct SHA-1, avoids having
                        # to decompress and calculate the SHA-1
                        sha1 = real_name
                    else:
                        # Assume we are not interested in the parent's SHA-1
                        sha1 = "ffffffffffffffffffffffffffffffffffffffff"

        if sha1 is None:
            if f is None:
                f = archive.open(path)
            while True:
                if self.stop_check():
                    return
                data = f.read(65536)
                if not data:
                    break
                raw_sha1_obj.update(data)
                if filter_name.startswith("Skip("):
                    filter_sha1_obj.update(data)
                    filter_size += len(data)
                elif filter_name == "ByteSwapWords":
                    # We don't really expect odd number of bytes when
                    # byteswapping words, but this handles the cases where we
                    # have "false positives" that aren't supposed to be
                    # byteswapped, and this prevents errors.
                    data_size = len(data)
                    for i in range(0, len(data), 2):
                        if data_size - i >= 2:
                            filter_sha1_obj.update(data[i + 1 : i + 2])
                            filter_sha1_obj.update(data[i : i + 1])
                            filter_size += 2
                        else:
                            filter_sha1_obj.update(data[i : i + 1])
                            filter_size += 1
            sha1 = raw_sha1_obj.hexdigest()
            filter_sha1 = filter_sha1_obj.hexdigest()

        if ext == ".rom":
            try:
                filter_data = ROMManager.decrypt_archive_rom(archive, path)
                filter_sha1 = filter_data["sha1"]
                filter_size = len(filter_data["data"])
            except Exception:
                import traceback

                traceback.print_exc()
                filter_sha1 = None
            if filter_sha1:
                if filter_sha1 != sha1:
                    print(
                        "[Files] Found encrypted rom {0} => {1}".format(
                            sha1, filter_sha1
                        )
                    )
                    # sha1 is now the decrypted sha1, not the actual sha1 of the
                    # file itself, a bit ugly, since md5 and crc32 are still
                    # encrypted hashes, but it works well with the kickstart
                    # lookup mechanism
                    # FIXME: Enable use of filter mechanism for Cloanto ROMs
                    sha1 = filter_sha1
                    # filter_name = "Cloanto"
            else:
                # If the ROM was encrypted and could not be decrypted, we
                # don't add it to the database. This way, the file will be
                # correctly added on later scans if rom.key is added to the
                # directory.
                return None

        if parent:
            path = "#/" + path.rsplit("#/", 1)[1]
        if parent is not None:
            # FIXME: size is incorrect here -- it's the size of the parent
            # file currently (not a big problem), setting it to -1 instead
            # for now.
            size = -1
        file_id = database.add_file(
            path=path, sha1=sha1, mtime=mtime, size=size, parent=parent
        )
        self.files_added += 1
        if parent is None:
            self.bytes_added += size

        if filter_name:
            if parent:
                # We want to add to the previous archive path
                pass
            else:
                # Reset path
                path = ""
            path += "#?Filter=" + filter_name
            # If not already in an archive (has a real file parent), set
            # parent to the real file
            database.add_file(
                path=path,
                sha1=filter_sha1,
                mtime=mtime,
                size=filter_size,
                parent=(parent or file_id),
            )

        if ext == ".rom":
            if self.on_rom_found:
                self.on_rom_found(path, sha1)
        return file_id


def check_valid_name(name):
    # check that the file is actually unicode object (indirectly). listdir
    # will return str objects for file names on Linux with invalid encoding.
    # FIXME: not needed for Python 3
    try:
        str(name)
    except UnicodeDecodeError:
        return False
    return True



import os
import shutil
import sys

from build.lib import (
    arch_name,
    macos,
    os_name,
    run,
    update_version_command,
    windows,
    # rmtree_if_exists,
    getPackageInformation,
    getAppleCodesignIdentity,
    getBundlePath,
    buildDmg,
    signDmg,
    notarizeApp,
    notarizeDmg,
    shell,
)


def version():
    update_version_command()
    info = getPackageInformation()
    with open("launcher/version.py", "w") as f:
        f.write(f'VERSION = "{info.version}"\n')
        f.write('PACKAGER = ""\n')
        f.write('COMMIT = ""\n')


def bootstrap():
    run(["python", "./bootstrap"])


def configure():
    pass


def _remove_unwanted_pyqt_files(dir):
    for name in [
        # Linux
        "libgcc_s.so.1",
        "libQt6Pdf.so.6",
        "libQt6Svg.so.6",
        "libstdc++.so.6",
        "PyQt6/Qt6/lib/libQt6Pdf.so.6",
        "PyQt6/Qt6/lib/libQt6Svg.so.6",
        # Windows
        # "d3dcompiler_47.dll",
        # "libcrypto-1_1-x64.dll",
        # "libEGL.dll",
        # "libgcc_s_seh-1.dll",
        # "libGLESv2.dll",
        # "libssl-1_1-x64.dll",
        # "libstdc++-6.dll",
        # "libwinpthread-1.dll",
        # "opengl32sw.dll",
        # "PyQt6/Qt/plugins/bearer/qgenericbearer.dll",
        # "PyQt6/Qt/plugins/imageformats/qicns.dll",
        # "PyQt6/Qt/plugins/imageformats/qico.dll",
        # "PyQt6/Qt/plugins/imageformats/qtga.dll",
        # "PyQt6/Qt/plugins/imageformats/qtiff.dll",
        # "PyQt6/Qt/plugins/imageformats/qwbmp.dll",
        # "PyQt6/Qt/plugins/platforms/qminimal.dll",
        # "PyQt6/Qt/plugins/platforms/qoffscreen.dll",
        # "PyQt6/Qt/plugins/platforms/qwebgl.dll",
        # "PyQt6/QtNetwork.pyd",
        # "PyQt6/QtQml.pyd",
        # "PyQt6/QtQuick.pyd",
        # "Qt63DAnimation.dll",
        # "Qt63DCore.dll",
        # "Qt63DInput.dll",
        # "Qt63DLogic.dll",
        # "Qt63DQuickScene2D.dll",
        # "Qt63DRender.dll",
        # "Qt6Bluetooth.dll",
        # "Qt6Concurrent.dll",
        # "Qt6DBus.dll",
        # "Qt6Gamepad.dll",
        # "Qt6Location.dll",
        # "Qt6Multimedia.dll",
        # "Qt6MultimediaQuick.dll",
        # "Qt6Network.dll",
        # "Qt6Nfc.dll",
        # "Qt6Positioning.dll",
        # "Qt6PositioningQuick.dll",
        # "Qt6Qml.dll",
        # "Qt6QmlModels.dll",
        # "Qt6QmlWorkerScript.dll",
        # "Qt6Quick.dll",
        # "Qt6Quick3D.dll",
        # "Qt6Quick3DAssetImport.dll",
        # "Qt6Quick3DRender.dll",
        # "Qt6Quick3DRuntimeRender.dll",
        # "Qt6Quick3DUtils.dll",
        # "Qt6QuickControls2.dll",
        # "Qt6QuickParticles.dll",
        # "Qt6QuickShapes.dll",
        # "Qt6QuickTemplates2.dll",
        # "Qt6QuickTest.dll",
        # "Qt6RemoteObjects.dll",
        # "Qt6Sensors.dll",
        # "Qt6Sql.dll",
        # "Qt6Test.dll",
        # "Qt6WebChannel.dll",
        # "Qt6WebSockets.dll",
        # "Qt6WebView.dll",
        # "Qt6XmlPatterns.dll",
        # macOS
    ]:
        path = os.path.join(dir, name)
        if os.path.exists(path):
            os.remove(path)


def build():
    # Running make to create locales (among other things)
    run(["make"])

    build_dir = "build/_build"
    script_name = "fs-uae-launcher"

    dist_dir = f"{build_dir}/pyinstaller"
    if os.path.exists(dist_dir):
        shutil.rmtree(dist_dir)
    args = [
        "pyinstaller",
        "--specpath",
        "build/_build/spec",
        "--distpath",
        f"{dist_dir}",
        "--log-level",
        "DEBUG",
        "--collect-submodules",
        "OpenGL",
    ]
    if windows or macos:
        args.append("--windowed")
    if windows:
        args.extend(["--icon", "../../../icon/fs-uae-launcher.ico"])
    if macos:
        args.extend(
            ["--osx-bundle-identifier", "no.fengestad.fs-uae-launcher"]
        )
    args.append(script_name)
    run(args)
    internal_dir = os.path.join(dist_dir, "fs-uae-launcher", "_internal")

    _remove_unwanted_pyqt_files(os.path.join(internal_dir))


def bundle():
    # package_name = "fs-uae-launcher"
    # package_name_pretty = "FS-UAE-Launcher"
    # package_display_name = "FS-UAE Launcher"
    # macos_bundle_id = "no.fengestad.fs-uae-launcher"
    package = getPackageInformation()

    if macos:
        src_dir = f"build/_build/pyinstaller/{package.name}.app"
        app_dir = f"build/_build/{package.display_name}.app"
        if os.path.exists(app_dir):
            shutil.rmtree(app_dir)

        shutil.move(src_dir, app_dir)

        # os.makedirs(bin_bin)
        contents_dir = os.path.join(app_dir, "Contents")
        # os.makedirs(contents_dir)
        info_plist_file = os.path.join(contents_dir, "Info.plist")
        with open(info_plist_file, "w") as f:
            f.write(f"""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
  <dict>
    <key>CFBundleDisplayName</key>
    <string>{package.display_name}</string>
    <key>CFBundleExecutable</key>
    <string>{package.name}</string>
    <key>CFBundleIconFile</key>
    <string>{package.name}.icns</string>
    <key>CFBundleIdentifier</key>
    <string>{package.macos_bundle_id}</string>
    <key>CFBundleInfoDictionaryVersion</key>
    <string>6.0</string>
    <key>CFBundleName</key>
    <string>{package.display_name}</string>
    <key>CFBundlePackageType</key>
    <string>APPL</string>
    <key>CFBundleShortVersionString</key>
    <string>{package.simple_version}</string>
    <key>NSRequiresAquaSystemAppearance</key>
    <false/>
    <key>NSHighResolutionCapable</key>
    <true/>
  </dict>
</plist>
""")
        resources_dir = os.path.join(contents_dir, "Resources")
        shutil.copy(
            "icon/fs-uae-launcher.icns",
            os.path.join(resources_dir, "fs-uae-launcher.icns"),
        )

        locale_dir = os.path.join(resources_dir, "Locale")
    else:
        bundle_dir = f"build/_build/{package.pretty_name}"
        if os.path.exists(bundle_dir):
            shutil.rmtree(bundle_dir)
        bin_dir = os.path.join(bundle_dir, os_name, arch_name)

        src = f"build/_build/pyinstaller/{package.name}"
        # shutil.copytree(src, bin_dir, symlinks=True)
        shutil.move(src, bin_dir)

        resources_dir = os.path.join(bundle_dir, "Resources")
        os.makedirs(resources_dir)

        # Or put locale_dir in Resources?
        locale_dir = os.path.join(bundle_dir, "Locale")

    shutil.copytree("share/locale", locale_dir)

    for item in ["arcade", "fsgs", "launcher", "workspace"]:
        archive = shutil.make_archive(
            os.path.join(resources_dir, item),
            "zip",
            os.path.join("share", "fs-uae-launcher", "resources", item),
        )
        print(archive)


def sign():
    args = [
        "codesign",
        "--force",
        "--deep",
        "--options",
        "runtime",
        "--sign",
        getAppleCodesignIdentity(),
        "--digest-algorithm=sha1,sha256",
    ]
    if os.path.exists("build/Entitlements.plist"):
        args.extend(["--entitlements", "build/Entitlements.plist"])
    args.append(getBundlePath())
    run(args)
    # runCodeSign(args)


def notarize():
    bundleId = getPackageInformation().macos_bundle_id
    bundlePath = getBundlePath()
    bundleName = os.path.basename(bundlePath)
    bundleParentDir = os.path.dirname(bundlePath)
    if os.path.exists("build/_build/notarize.zip"):
        os.remove("build/_build/notarize.zip")
    # zip = "../../../notarize.zip"
    zip = "notarize.zip"
    shell(
        f'cd {bundleParentDir} && ditto -c -k --keepParent "{bundleName}" "{zip}"'
    )
    notarizeApp("build/_build/notarize.zip", bundleId)
    # if bundlePath.endswith(".framework"):
    #     print(
    #         "Does not seem to be possible to staple tickets to frameworks? (error 73)"
    #     )
    #     print("Exiting...")
    #     sys.exit(0)
    run(["xcrun", "stapler", "staple", bundlePath])


def archive():
    package = getPackageInformation()
    archive_base_name = (
        f"{package.pretty_name}_{package.version}_" + f"{os_name}_{arch_name}"
    )
    if windows:
        # archive_file = archive_base_name + ".zip"
        archive_format = "zip"
    elif macos:
        archive_format = ""
    else:
        archive_format = "xztar"

    if archive_format:
        archive = shutil.make_archive(
            os.path.join("build", "_dist", archive_base_name),
            archive_format,
            os.path.join("build", "_build"),
            package.pretty_name,
        )
        print(archive)


def build_dmg():
    buildDmg()


def sign_dmg():
    signDmg()


def notarize_dmg():
    notarizeDmg()


def source():
    run(["make", "dist-xz"])
    package = getPackageInformation()
    archive = f"{package.name}-{package.version}.tar.xz"
    if not os.path.exists("build/_dist"):
        os.makedirs("build/_dist")
    shutil.move(archive, "build/_dist")


def default():
    bootstrap()
    configure()
    build()
    bundle()


def all():
    # version()
    bootstrap()
    configure()
    build()
    bundle()
    if macos:
        sign()
        notarize()
        build_dmg()
        sign_dmg()
        notarize_dmg()
    else:
        archive()


def main():
    args = sys.argv[1:]
    if len(args) == 0:
        args = ["default"]
    for arg in args:
        command = arg
        if command.startswith("--"):
            continue
        print("BUILD command:", command)
        if command == "all":
            all()
        elif command == "default":
            default()
        elif command == "version":
            version()
        elif command == "bootstrap":
            bootstrap()
        elif command == "configure":
            configure()
        elif command == "build":
            build()
        elif command == "bundle":
            bundle()
        elif command == "sign":
            sign()
        elif command == "notarize":
            notarize()
        elif command == "archive":
            archive()
        elif command == "build-dmg":
            build_dmg()
        elif command == "sign-dmg":
            sign_dmg()
        elif command == "notarize-dmg":
            notarize_dmg()
        elif command == "source":
            source()
        else:
            print(f"Unsupported command: {command}", file=sys.stderr)
            sys.exit(1)


if __name__ == "__main__":
    main()



from fsui.qt import QWidget


# noinspection PyPep8Naming
def QParent(parent, window=False):
    if parent is None:
        return None
    if window:
        if hasattr(parent, "real_window"):
            return parent.real_window()
    else:
        if hasattr(parent, "real_widget"):
            return parent.real_widget()
    if hasattr(parent, "get_container"):
        return parent.get_container()
    if isinstance(parent, QWidget):
        return parent
    raise Exception("Could not find QParent")



import json
import traceback

from arcade.glui.opengl import fs_emu_blending, fs_emu_texturing
from OpenGL import GL as gl
from OpenGL.GLU import gluOrtho2D
from arcade.glui.texture import Texture
from arcade.resources import resources
from fsbc.util import memoize
from fsui.qt import (
    QColor,
    QFontDatabase,
    QFontMetrics,
    QImage,
    QPainter,
    QPen,
    QPoint,
    is_pyqt6,
)

CACHE_SIZE = 100
text_cache = []
for _ in range(CACHE_SIZE):
    text_cache.append({"text": None, "font": None, "texture": None})


class Font(object):
    title_font = None
    subtitle_font = None
    small_font = None
    main_path_font = None
    list_subtitle_font = None
    header_font = None

    font_ids = {}

    def __init__(self, path, size):
        if is_pyqt6:
            self.database = QFontDatabase
        else:
            self.database = QFontDatabase()
        try:
            self.font_id = Font.font_ids[path]
        except Exception:
            if isinstance(path, str):
                self.font_id = self.database.addApplicationFont(path)
            else:
                self.font_id = self.database.addApplicationFontFromData(
                    path.read()
                )
            Font.font_ids[path] = self.font_id
        self.families = self.database.applicationFontFamilies(self.font_id)
        print(self.families)
        self.size = size
        if len(self.families) > 0:
            styles = self.database.styles(self.families[0])
            print(styles)
            self.font = self.database.font(self.families[0], "Bold", size)
            if self.size <= 0:
                print("WARNING: font size is", self.size)
                traceback.print_stack()
            self.font.setPixelSize(self.size)
        else:
            self.font = None

    def set_size(self, size):
        if self.font is not None:
            self.size = size
            if self.size <= 0:
                print("WARNING: font size is", self.size)
                traceback.print_stack()
            self.font.setPixelSize(self.size)

    def render(self, text, _, color):
        if self.font is None:
            return "", (0, 0)

        fm = QFontMetrics(self.font)
        rect = fm.boundingRect(text)
        im = QImage(
            rect.x() + rect.width(),
            rect.height(),
            QImage.Format.Format_ARGB32_Premultiplied,
        )
        im.fill(QColor(0, 0, 0, 0))
        painter = QPainter()
        painter.begin(im)
        painter.setPen(QPen(QColor(*color)))
        painter.setFont(self.font)
        painter.drawText(QPoint(0 - rect.x(), 0 - rect.y()), text)
        painter.end()

        bits = im.bits()
        try:
            pixels = bits.tobytes()
        except AttributeError:
            bits.setsize(im.sizeInBytes())
            pixels = bytes(bits)
        return pixels, (rect.x() + rect.width(), rect.height())

    def rendered_size(self, text):
        if self.font is None:
            return 0, 0

        fm = QFontMetrics(self.font)
        rect = fm.boundingRect(text)
        return rect.width(), rect.height()


class BitmapFont(object):
    title_font = None
    menu_font = None

    def __init__(self, name):
        self.texture = None
        self.x = []
        self.y = []
        self.w = [0 for _ in range(256)]
        self.h = 0
        self.load(name)

    @memoize
    def measure(self, text):
        required_width = 0
        required_height = self.h
        for c in text:
            required_width += self.w[ord(c)]
        return required_width, required_height

    def render(self, text, x, y, r=1.0, g=1.0, b=1.0, alpha=1.0):
        global text_cache
        if not text:
            return 0

        # find cached text entry, if any

        for i, item in enumerate(text_cache):
            if item["text"] == text and item["font"] == self:
                text_cache.pop(i)
                break
        else:
            item = None

        if item:
            fs_emu_blending(True)
            fs_emu_texturing(True)

            w = item["w"]
            h = item["h"]
            gl.glBindTexture(gl.GL_TEXTURE_2D, item["texture"])
            gl.glBegin(gl.GL_QUADS)
            gl.glColor4f(r * alpha, g * alpha, b * alpha, alpha)
            gl.glTexCoord2d(0.0, 0.0)
            gl.glVertex2d(x, y)
            gl.glTexCoord2d(1.0, 0.0)
            gl.glVertex2d(x + w, y)
            gl.glTexCoord2d(1.0, 1.0)
            gl.glVertex2d(x + w, y + h)
            gl.glTexCoord2d(0.0, 1.0)
            gl.glVertex2d(x, y + h)
            gl.glEnd()

            # re-insert item at front
            text_cache.insert(0, item)
            return w, h

        # calculate size of text

        required_width, required_height = self.measure(text)

        # setup fbo

        render_texture = gl.glGenTextures(1)
        gl.glBindTexture(gl.GL_TEXTURE_2D, render_texture)
        gl.glTexImage2D(
            gl.GL_TEXTURE_2D,
            0,
            gl.GL_RGBA,
            required_width,
            required_height,
            0,
            gl.GL_RGBA,
            gl.GL_UNSIGNED_INT,
            None,
        )
        gl.glTexParameteri(
            gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S, gl.GL_CLAMP_TO_EDGE
        )
        gl.glTexParameteri(
            gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T, gl.GL_CLAMP_TO_EDGE
        )
        gl.glTexParameteri(
            gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR
        )

        # FIXME: Mipmapping?
        mip_mapping = 0
        if mip_mapping:
            gl.glTexParameteri(
                gl.GL_TEXTURE_2D,
                gl.GL_TEXTURE_MIN_FILTER,
                gl.GL_LINEAR_MIPMAP_LINEAR,
            )
            gl.glTexParameteri(
                gl.GL_TEXTURE_2D, gl.GL_GENERATE_MIPMAP, gl.GL_TRUE
            )
            gl.glGenerateMipmap(gl.GL_TEXTURE_2D)
        else:
            gl.glTexParameteri(
                gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR
            )

        gl.glBindTexture(gl.GL_TEXTURE_2D, 0)

        # Set up some renderbuffer state

        from OpenGL.GL import (
            glBindFramebuffer,
            glCheckFramebufferStatus,
            glDeleteFramebuffers,
            glFramebufferTexture2D,
            glGenFramebuffers,
        )

        frame_buffer = gl.GLuint()
        glGenFramebuffers(1, frame_buffer)
        glBindFramebuffer(gl.GL_FRAMEBUFFER, frame_buffer)
        glFramebufferTexture2D(
            gl.GL_FRAMEBUFFER,
            gl.GL_COLOR_ATTACHMENT0,
            gl.GL_TEXTURE_2D,
            render_texture,
            0,
        )

        status = glCheckFramebufferStatus(gl.GL_FRAMEBUFFER)
        if status != gl.GL_FRAMEBUFFER_COMPLETE:
            print("glCheckFramebufferStatusEXT error", status)

        gl.glPushMatrix()
        gl.glLoadIdentity()
        gl.glPushAttrib(int(gl.GL_VIEWPORT_BIT) | int(gl.GL_ENABLE_BIT))
        gl.glViewport(0, 0, required_width, required_height)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPushMatrix()
        gl.glLoadIdentity()
        gluOrtho2D(0, required_width, 0, required_height)

        gl.glClearColor(0.0, 0.0, 0.0, 0.0)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT)

        gl.glEnable(gl.GL_BLEND)
        gl.glEnable(gl.GL_TEXTURE_2D)

        self.texture.bind()

        tw = self.texture.w
        th = self.texture.h

        gl.glBegin(gl.GL_QUADS)
        gl.glColor4f(1.0, 1.0, 1.0, 1.0)
        x2 = 0
        h = self.h
        for ch in text:
            c = ord(ch)
            w = self.w[c]
            s1 = self.x[c] / tw
            s2 = (self.x[c] + w) / tw
            t1 = (self.y[c]) / th
            t2 = (self.y[c] + h) / th
            gl.glTexCoord2d(s1, t2)
            gl.glVertex2d(x2, 0)
            gl.glTexCoord2d(s2, t2)
            gl.glVertex2d(x2 + w, 0)
            gl.glTexCoord2d(s2, t1)
            gl.glVertex2d(x2 + w, h)
            gl.glTexCoord2d(s1, t1)
            gl.glVertex2d(x2, h)
            x2 += w
        gl.glEnd()

        gl.glPopMatrix()
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glPopAttrib()
        glBindFramebuffer(gl.GL_FRAMEBUFFER, 0)
        gl.glPopMatrix()

        glDeleteFramebuffers(1, frame_buffer)

        if mip_mapping:
            gl.glBindTexture(gl.GL_TEXTURE_2D, render_texture)
            gl.glGenerateMipmap(gl.GL_TEXTURE_2D)
            gl.glBindTexture(gl.GL_TEXTURE_2D, 0)

        new_item = {
            "font": self,
            "text": text,
            "w": required_width,
            "h": required_height,
            "texture": render_texture,
        }
        text_cache.insert(0, new_item)

        item = text_cache.pop()
        if item["texture"]:
            gl.glDeleteTextures([item["texture"]])

        # now the text is in the cache, so call function again
        return self.render(text, x, y, r, g, b, alpha)

    def load(self, name):
        print("load font", name)
        self.texture = Texture(name + ".png")
        f = resources.resource_stream(name + ".json")
        doc = json.loads(f.read().decode("UTF-8"))
        self.x = doc["x"]
        self.y = doc["y"]
        self.w = doc["w"]
        self.h = doc["h"][0]
